# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import requests
import pickle
import requests.status_codes
import os
from dashboard_cli import __version__


class ApiClient:
    def __init__(self, base_url=None, timeout=30, session_file="~/.dashboard_session.bin", verbose=False, verify=True):
        self.base_url = base_url
        self.session_file = os.path.expanduser(session_file)
        self.timeout = timeout
        self.verbose = verbose
        self.verify = verify

        if not self.verify:  # pragma: no cover
            self.debug("WARNING: Insecure mode is activated: TLS/SSL certificate verification is disabled.")

        self.load_or_create_session()

    def debug(self, *what):
        if self.verbose:  # pragma: no cover
            print(" ".join((map(str, what))))

    def load_or_create_session(self):
        try:
            with open(self.session_file, "rb") as fp:
                self.session = pickle.load(fp)
                self.debug("loaded session with cookies from {}".format(self.session_file))
        except Exception:
            self.session = requests.Session()

        self.session.headers.update({"User-Agent": f"dashboard-cli-{__version__}"})

    def persist_session(self):
        # save session
        self.debug("saved session with cookies to {}".format(self.session_file))

        with open(self.session_file, "wb") as fp:
            pickle.dump(self.session, fp)

    def build_url(self, api_endpoint):
        """
        Concatenate url which is an API endpoint and the current server url.
        """
        assert not api_endpoint.lower().startswith("http:"), f"url ({api_endpoint}) must not start with a scheme!"
        assert api_endpoint.startswith("/")
        url = f"{self.base_url}{api_endpoint}"
        return url

    def _is_response_successful(self, response):
        return 200 <= response.status_code < 300

    def get(self, url):
        url = self.build_url(url)
        self.debug(f"GET {url} timeout={self.timeout}")
        response = self.session.get(url, timeout=self.timeout, verify=self.verify)
        self.debug(f"Response: {response}")

        if response is not None:  # pragma: no branch
            if self._is_response_successful(response):
                self.persist_session()

        return response

    def post(self, url, data):
        url = self.build_url(url)
        self.debug(f"POST {url} data={data} timeout={self.timeout}")
        response = self.session.post(url, json=data, timeout=self.timeout, verify=self.verify)
        self.debug(f"Response: {response}")

        if response is not None:  # pragma: no branch
            if self._is_response_successful(response):
                self.persist_session()

        return response

    def put(self, url, data):
        url = self.build_url(url)
        self.debug(f"POST {url} data={data} timeout={self.timeout}")
        response = self.session.put(url, json=data, timeout=self.timeout, verify=self.verify)
        self.debug(f"Response: {response}")

        if response is not None:  # pragma: no branch
            if self._is_response_successful(response):
                self.persist_session()

        return response

    def delete(self, url, data=None):
        url = self.build_url(url)
        self.debug(f"POST {url} data={data} timeout={self.timeout}")
        response = self.session.delete(url, json=data, timeout=self.timeout, verify=self.verify)
        self.debug(f"Response: {response}")

        if response is not None:  # pragma: no branch
            if self._is_response_successful(response):  # pragma: no branch
                self.persist_session()

        return response

    def login(self, username=None, password=None):
        self.session.auth = None
        if username is not None:  # pragma: no branch
            if password is not None:  # pragma: no branch
                self.session.auth = (username, password)

        response = self.post(
            "/api/v1/login",
            data={'username': username, 'password': password}
        )

        if response is not None:  # pragma: no branch
            if response.status_code == 403:
                raise Exception("Wrong username or password!")
            else:
                assert response.status_code == 204

        return response

    def logout(self):
        response = self.post(
            "/api/v1/logout",
            data=None
        )

        if response is not None:  # pragma: no branch
            if response.status_code == 403:
                raise Exception("Login first!")
            else:
                assert response.status_code == 204

        os.unlink(self.session_file)

        return response
