# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, HILSTER GmbH, Germany - https://www.hilster.io
# All rights reserved.
#

import getpass
import json
import os
import sys
import zipfile
import argcomplete
import argparse
import tabulate
import time
import re

from dashboard_cli.apiclient import ApiClient


def project(project_name):
    """
    Project name validator. 
    """
    if re.match(r"^[A-Za-z0-9-_]+$", project_name) is None:
        raise ValueError(f"{project} is not a valid project identifier. It must only contain A-Z, a-z, 0-9 - or _.")

    return project_name


class DashboardCLI:
    def __init__(self, parse_args=True):
        self.timeout = 30
        if parse_args is True:  # pragma: no branch
            self.parse_arguments(sys.argv)

        self.client = ApiClient(
            base_url=self.server,
            verbose=self.verbose,
            session_file=self.session_file,
            timeout=self.timeout,
            verify=not self.insecure
        )

        if parse_args is True:  # pragma: no branch
            self.handle_command()

    def handle_command(self):
        if self.server_username is not None:
            if self.command != "login":
                self.login()
        if self.command == "login":
            self.login()
        elif self.command == "logout":
            self.logout()
        elif self.command == "add-user":
            self.add_user()
        elif self.command == "delete-user":
            self.delete_user()
        elif self.command == "user-info":
            self.user_info()
        elif self.command == "whoami":
            self.whoami()
        elif self.command == "passwd":
            self.passwd()
        elif self.command == "list-users":
            self.list_users()
        elif self.command == "set-roles":
            self.set_roles()
        elif self.command == 'rebuild-cache':
            self.rebuild_cache()
        elif self.command == 'list-projects':
            self.list_projects()
        elif self.command == 'status':
            self.status()
        elif self.command == 'list-reports':
            self.list_reports()
        elif self.command == 'delete-report':
            self.delete_report()
        elif self.command == 'list-requirements':
            self.list_requirements()
        elif self.command == 'delete-requirements':
            self.delete_requirements()
        elif self.command == 'create-project':
            self.create_project()
        elif self.command == 'delete-project':
            self.delete_project()
        elif self.command == 'export-project':
            self.export_project()
        elif self.command == 'import-project':
            self.import_project()
        elif self.command == 'upload-requirements':
            self.upload_requirements()
        elif self.command == 'upload-reports':
            self.upload_reports()
        elif self.command == 'create-token':
            self.create_token()
        elif self.command == 'delete-token':
            self.delete_token()
        elif self.command == 'list-tokens':
            self.list_tokens()
        elif self.command == 'delete-tokens':  # pragma: no branch
            self.delete_tokens()
        elif self.command == 'complete':  # pragma: no cover
            print("eval \"$(register-python-argcomplete dashboard)\"")
        else:  # pragma: no cover
            raise Exception(f'Unknown command: {self.command}')

    def parse_arguments(self, argv):
        self._init_argument_parsers()
        self._main_parser.parse_args(argv[1:], self)

    def _init_argument_parsers(self):
        self._main_parser = argparse.ArgumentParser(
            description="QABench Dashboard CLI - https://qabench.io"
        )

        self._main_parser.add_argument("-s", "--server", dest="server", help="Server with HTTP URI scheme (default: http://localhost)",
                                       default=os.getenv("DASHBOARD_SERVER", "http://localhost"))

        self._main_parser.add_argument("-u", "--username", dest="server_username", help="Username",
                                       default=os.getenv("DASHBOARD_USERNAME", None))

        self._main_parser.add_argument("-p", "--password", dest="server_password", help="Password",
                                       default=os.getenv("DASHBOARD_PASSWORD", None))

        self._main_parser.add_argument("-t", "--timeout", dest="timeout", help="Timeout in seconds (default: 30 s)",
                                       type=int, default=os.getenv("DASHBOARD_TIMEOUT", 30))

        self._main_parser.add_argument("-v", "--verbose", dest="verbose", help="Verbose output",
                                       action="store_true",
                                       default=bool(os.getenv("DASHBOARD_VERBOSE", False)))

        self._main_parser.add_argument("-S", "--session-file",
                                       dest="session_file", help="Filename to store session information (default: ~/.dashboard_session.bin)",
                                       default=os.getenv("DASHBOARD_SESSION_FILE", "~/.dashboard_session.bin"))

        self._main_parser.add_argument("--insecure", dest="insecure", help="Do not check TLS/SSL cerfiticates",
                                       action="store_true",
                                       default=bool(os.getenv("DASHBOARD_INSECURE", False)))

        self._command = self._main_parser.add_subparsers(
            dest="command", help=""
        )
        self._command.required = True

        self._login_parser = self._command.add_parser("login", help="Login")

        self._logout_parser = self._command.add_parser("logout", help="Logout")

        self._add_user_parser = self._command.add_parser("add-user", help="Add a user")
        self._add_user_parser.add_argument("username", help="username")
        self._add_user_parser.add_argument("--password", dest="password", default=None, help="password")

        self._delete_user_parser = self._command.add_parser("delete-user", help="Add a user")
        self._delete_user_parser.add_argument("username", help="username")

        self._passwd_parser = self._command.add_parser("passwd", help="Change password")
        self._passwd_parser.add_argument("--username", default=None, help="username")
        self._passwd_parser.add_argument("--password", default=None, help="password")

        self._user_info_parser = self._command.add_parser("user-info", help="Get user information")
        self._user_info_parser.add_argument("--username", default=None, help="username")

        self._whoami_parser = self._command.add_parser("whoami", help="Get current username")

        self._set_roles_parser = self._command.add_parser("set-roles", help="Set user's roles for a project")
        self._set_roles_parser.add_argument(dest="project", type=project, help="Project")
        self._set_roles_parser.add_argument("user", help="username")
        self._set_roles_parser.add_argument("roles", nargs="*", help="roles (may be repeated)")

        self._list_users_parser = self._command.add_parser("list-users", help="List users")
        self._list_users_parser.add_argument("--user", default=None, help="username")

        self._create_project_parser = self._command.add_parser("create-project", help="Create a project")
        self._create_project_parser.add_argument(dest="project", type=project, help="Project name")
        self._create_project_parser.add_argument(dest="project_alias", help="Human readable name")

        self._delete_project_parser = self._command.add_parser("delete-project", help="Delete a project")
        self._delete_project_parser.add_argument(dest="project", type=project, help="Project")

        self._list_projects_parser = self._command.add_parser("list-projects", help="Show a list of projects")

        self._project_status_parser = self._command.add_parser("status", help="Show status of a project")
        self._project_status_parser.add_argument(dest="project", type=project, help="Project name")

        self._list_reports_parser = self._command.add_parser("list-reports", help="Show a list of reports")
        self._list_reports_parser.add_argument(dest="project", type=project, help="Project")

        self._delete_reports_parser = self._command.add_parser("delete-report", help="Delete a report")
        self._delete_reports_parser.add_argument(dest="project", type=project, help="Project")
        self._delete_reports_parser.add_argument(dest="report_id", help="Report ID")

        self._list_requirements_parser = self._command.add_parser("list-requirements", help="Show a list of requirements")
        self._list_requirements_parser.add_argument(dest="project", type=project, help="Project")

        self._delete_requirements_parser = self._command.add_parser("delete-requirements", help="Delete requirements")
        self._delete_requirements_parser.add_argument(dest="project", type=project, help="Project")
        self._delete_requirements_parser.add_argument(dest="requirements_id", help="Requirements ID")

        self._create_token_parser = self._command.add_parser("create-token", help="Create a token")
        self._create_token_parser.add_argument(dest="project", type=project, help="Project")
        self._create_token_parser.add_argument(dest="name", help="Name of the token")

        self._delete_token_parser = self._command.add_parser("delete-token", help="Delete a token")
        self._delete_token_parser.add_argument(dest="project", type=project, help="Project")
        self._delete_token_parser.add_argument(dest="token", help="The token to deleted (name or hash)")

        self._list_tokens_parser = self._command.add_parser("list-tokens", help="List project tokens")
        self._list_tokens_parser.add_argument(dest="project", type=project, help="Project")

        self._delete_tokens_parser = self._command.add_parser("delete-tokens", help="Delete all project tokens")
        self._delete_tokens_parser.add_argument(dest="project", type=project, help="Project")

        self._export_parser = self._command.add_parser("export-project", help="Export a project to an archive")
        self._export_parser.add_argument(dest="project", type=project, help="Project")
        self._export_parser.add_argument(dest="project_archive", help="Project Archive")

        self._import_parser = self._command.add_parser("import-project", help="Import a project from an archive")
        self._import_parser.add_argument('-c', dest="create_on_import", action='store_true',
                                         help="Create project if necessary", default=False)
        self._import_parser.add_argument("-w", "--wait", dest="wait", type=int, default=0,
                                        help="Wait until cache is updated (can be used with --timeout option, too)")
        self._import_parser.add_argument(dest="project", type=project, help="Project")
        self._import_parser.add_argument(dest="project_archive", help="Project Archive")

        # dashboard-cli upload-requirements <server-url> <requirements>
        self._upload_requirements_parser = self._command.add_parser("upload-requirements", help="Upload requirements")
        self._upload_requirements_parser.add_argument(dest="project", type=project, help="Project")
        self._upload_requirements_parser.add_argument(dest="requirements", nargs="+", help="Requirements")
        self._upload_requirements_parser.add_argument("--date",
                                                      help='Export date of the requirements (optional), '
                                                           'i.e. "2017-09-23 09:30:35"')

        # dashboard-cli upload-reports <server-url> <reports>
        self._upload_reports_parser = self._command.add_parser("upload-reports", help="Upload reports")
        self._upload_reports_parser.add_argument(dest="project", type=project, help="Project")
        self._upload_reports_parser.add_argument(dest="reports", nargs="+", help="Reports")

        self._rebuild_cache_parser = self._command.add_parser("rebuild-cache", help="Rebuild cache")
        self._rebuild_cache_parser.add_argument(dest="project", type=project, help="Project")
        self._rebuild_cache_parser.add_argument("-w", "--wait", dest="wait", type=int, default=0, help="Wait until cache is updated (can be used with --timeout option, too)")

        self._completeParser = self._command.add_parser('complete', help="print an evaluatable expression for bash completion "
                                                                            "(use with 'eval $(dashboard complete)')")

        argcomplete.autocomplete(self._main_parser)

    def login(self):
        if self.server_username is None:  # pragma: no cover
            self.server_username = input("Username: ")

        if self.server_password is None:  # pragma: no cover
            self.server_password = getpass.getpass("Password: ")

        self.client.login(username=self.server_username, password=self.server_password)

    def logout(self):
        try:
            self.client.logout()
        except FileNotFoundError:  # pragma: no cover
            pass

    def add_user(self):
        username = self.username
        password = self.password
        if not password:  # pragma: no cover
            # get password from command line
            password0 = getpass.getpass(f"Enter password for {username}: ")
            password1 = getpass.getpass(f"Enter password again: ")
            assert password0 == password1, "Passwords do not match!"

            password = password0

        url = "/api/v1/user/" + username

        ret = self.client.post(url, data={'password': password})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def delete_user(self):
        username = self.username
        url = "/api/v1/user/" + username

        ret = self.client.delete(url)
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def user_info(self):
        username = self.username
        url = "/api/v1/user"

        if username:  # pragma: no branch
            url += "/" + username

        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        data = ret.json()

        user = data["username"]
        roles = {}
        try:
            roles = data["roles"]
        except KeyError:
            pass

        self._print_user_and_roles(user, roles)

    def whoami(self):
        url = "/api/v1/user"

        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        print(ret.json()["username"])

    def passwd(self):
        username = self.username
        password = self.password
        if not password:  # pragma: no cover
            # get password from command line
            password0 = getpass.getpass(f"Enter new password: ")
            password1 = getpass.getpass(f"Enter new password again: ")
            assert password0 == password1, "Passwords do not match!"

            password = password0

        url = "/api/v1/user"

        if username:  # pragma: no branch
            url += "/" + username

        ret = self.client.put(url, data={'password': password})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def _print_user_and_roles(self, user, roles):
        print(f"{user}")
        for project, project_roles in sorted(roles.items()):
            print(f"    {project}: {', '.join(sorted(project_roles))}")

    def list_users(self):
        url = "/api/v1/users"

        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        for line in sorted(ret.json(), key=lambda x: x["username"]):
            user = line["username"]
            roles = {}
            try:
                roles = line["roles"]
            except KeyError:
                pass
            self._print_user_and_roles(user, roles)

    def set_roles(self):
        url = "/api/v1/user/" + self.user + "/roles"

        for role in self.roles:
            if role not in ["admin", "user"]:
                raise Exception("roles must be one of 'admin' or 'user'")

        ret = self.client.put(url, data={
            self.project: self.roles
        })

        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def create_project(self):
        print(f"Create project {self.project} with alias '{self.project_alias}'")

        url = "/api/v1/project/" + self.project

        ret = self.client.post(url, data={'alias': self.project_alias})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def delete_project(self):
        print(f"Delete project {self.project}")

        url = "/api/v1/project/" + self.project

        ret = self.client.delete(url, data={})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def list_projects(self):
        # print("project list handler")

        url = "/api/v1/projects"
        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        data = ret.json()
        projects = []
        aliases = []
        for project, doc in sorted(data.items()):
            projects.append(project)
            aliases.append(doc['alias'])

        number_of_reports = []
        for project in projects:
            reports = self.client.get(f"/api/v1/project/{project}/reports")
            assert reports.status_code == 200, f"Request returned an error: {ret}"
            n_reports = len(reports.json()['reports'])
            number_of_reports.append(n_reports)

        print_list = [list(a) for a in zip(projects, aliases, number_of_reports)]

        print(tabulate.tabulate(print_list, headers=['Project', 'Alias', 'Number of Reports']))

        return projects

    def status(self):
        project = self.client.get(f"/api/v1/project/{self.project}")
        assert project.status_code == 200, f"Request returned an error: {project}"
        json_project = project.json()
        tests_percentage = int(json_project['latest_number_of_successes'] * 100 / json_project['latest_number_of_tests'])
        coverage_percentage = int(json_project['latest_number_of_successfully_covered_requirements'] * 100 / json_project['latest_number_of_requirements'])

        print(f"""\
Project: {json_project['alias']}
Last modified: {json_project['last_modified']}
Successful tests: {tests_percentage} % ({json_project['latest_number_of_successes']}/{json_project['latest_number_of_tests']})
Requirements coverage: {coverage_percentage} % ({json_project['latest_number_of_successfully_covered_requirements']}/{json_project['latest_number_of_requirements']})\
""")

    def list_reports(self):
        ret = self.client.get(f"/api/v1/project/{self.project}/reports")
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        data = ret.json()

        print_list = [list((report["id"], report["date"], report["test_amount"],
                            f'Success {report["successes"]} Failure {report["failures"]} '
                            f'Error {report["errors"]} Skipped {report["skipped"]}')) for report in data['reports']]

        print(tabulate.tabulate(print_list, headers=['ID', 'Date', 'Number of Tests', 'Results']))

    def delete_report(self):
        # print("delete report handler")
        url = f"/api/v1/project/{self.project}/report/" + self.report_id
        ret = self.client.delete(url, data={})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def list_requirements(self):
        # print("list requirements handler")
        url = f"/api/v1/project/{self.project}/requirements"
        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        requirements = ret.json()

        print_list = [list((req["id"], req["date"], req["number_of_requirements"], req["requirements"])) for req in requirements]

        print(tabulate.tabulate(print_list, headers=['ID', 'Date', 'Number of Requirements', 'Requirements']))

    def delete_requirements(self):
        # print("delete requirements handler")
        url = f"/api/v1/project/{self.project}/requirements/" + self.requirements_id
        ret = self.client.delete(url, data={})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

    def create_token(self):
        url = f"/api/v1/project/{self.project}/token"
        json_data = {'name': self.name}
        ret = self.client.post(url, data=json_data)
        assert ret.status_code == 201, f"Request returned an error: {ret}"

        ret = ret.json()
        token = ret["token"]
        print(f"Created token '{self.name}': {token}")

    def delete_token(self, token=None):
        if token:
            self.token = token
        url = f"/api/v1/project/{self.project}/token/" + self.token
        ret = self.client.delete(url, data={})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

        print(f"Deleted token '{self.token}'")
        return self.token

    def _get_tokens(self):
        url = f"/api/v1/project/{self.project}/tokens"
        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        tokens = {value: key for key, value in ret.json().items()}
        return tokens

    def list_tokens(self):
        tokens = self._get_tokens()

        print_list = [[a[0], a[1]] for a in tokens.items()]

        print(tabulate.tabulate(print_list, headers=['Name', 'Token']))

        return tokens

    def delete_tokens(self):
        tokens = self._get_tokens()

        for token in tokens.values():
            self.delete_token(token)

    def upload_file(self, report, api_endpoint, date=None):
        with open(report, "r") as fp:
            json_data = {"filename": report,
                         "contents": fp.read(),
                         'date': date,
                         'type': None,
                         'origin': 'cli'}

            ret = self.client.post(api_endpoint, data=json_data)
            assert ret.status_code == 204, f"Request returned an error: {ret}"

    def upload_reports(self):
        api_endpoint = f"/api/v1/project/{self.project}/report"

        n = 1
        number_of_reports = len(self.reports)
        for report in self.reports:
            self.upload_file(report, api_endpoint)
            print("{}: {:>3.0f} % ({}/{})".format(report, 100.0 * n / number_of_reports, n, number_of_reports),
                  end="\r")
            n += 1
        print()

    def upload_requirements(self):
        api_endpoint = f"/api/v1/project/{self.project}/requirements"

        for requirements in self.requirements:
            print(f"Upload requirements {requirements} .. ", end="")
            self.upload_file(requirements, api_endpoint, date=self.date)
            print("OK")

    def _get_report_ids(self):
        url = f"/api/v1/project/{self.project}/reports"
        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        report_ids = [report['id'] for report in ret.json()['reports']]
        return report_ids

    def _get_requirement_ids(self):
        url = f"/api/v1/project/{self.project}/requirements"
        ret = self.client.get(url)
        assert ret.status_code == 200, f"Request returned an error: {ret}"

        requirement_ids = [req['id'] for req in ret.json()]
        return requirement_ids

    def export_project(self):
        export_report_endpoint = f"/api/v1/project/{self.project}/report"
        export_requirement_endpoint = f"/api/v1/project/{self.project}/requirements"
        report_ids = self._get_report_ids()
        requirement_ids = self._get_requirement_ids()
        tokens = self._get_tokens()

        number_of_reports = len(report_ids)
        number_of_requirements = len(requirement_ids)

        n = 1
        with zipfile.ZipFile(self.project_archive, 'w', compression=zipfile.ZIP_DEFLATED) as archive:
            print("Export Reports")
            print("==============")
            for report_id in report_ids:
                report_url = export_report_endpoint + f'/{report_id}'

                ret = self.client.get(report_url)
                assert ret.status_code == 200, f"Request returned an error: {ret}"

                report = ret.json()

                filename = os.path.join('reports', 'report_' + str(report['id']) + '.json')

                print("{}: {:>3.0f} % ({}/{})".format(filename, 100.0 * n / number_of_reports, n, number_of_reports),
                      end="\r")
                n += 1
                with archive.open(filename, 'w') as f:
                    f.write(json.dumps(report).encode())
            else:
                print("\n")

            n = 1
            print("Export Requirements")
            print("===================")
            for requirement_id in requirement_ids:
                requirement_url = export_requirement_endpoint + f'/{requirement_id}'

                ret = self.client.get(requirement_url)
                assert ret.status_code == 200, f"Request returned an error: {ret}"

                requirement = ret.json()

                filename = os.path.join('requirements', 'requirement_' + str(requirement_id) + '.json')

                print(
                    "{}: {:>3.0f} % ({}/{})".format(filename, 100.0 * n / number_of_requirements, n, number_of_requirements),
                    end="\r")
                n += 1
                with archive.open(filename, 'w') as f:
                    f.write(json.dumps(requirement).encode())
            else:
                print("\n")

            print("Export Tokens")
            print("=============")
            with archive.open('tokens.json', 'w') as f:
                f.write(json.dumps(tokens).encode())

        print()

    def import_project(self):
        if self.create_on_import:
            self.project_alias = self.project
            print("Create Project")
            print("==============")
            self.create_project()
            print()
        n = 1
        with zipfile.ZipFile(self.project_archive, 'r') as archive:
            reports = list(filter(lambda name: name.startswith('report') and name.endswith(".json"), archive.namelist()))
            number_of_reports = len(reports)
            print("Import Reports")
            print("==============")
            for report in reports:

                with archive.open(report, "r") as fp:
                    json_data = {
                        "filename": report.split('/')[-1],
                        "contents": fp.read().decode(),
                        'type': None,
                        'date': None,
                        'update_cache': False
                    }

                    ret = self.client.post(f"/api/v1/project/{self.project}/report",
                                           data=json_data)
                    assert ret.status_code == 204, f"Request returned an error: {ret}"
                print("{}: {:>3.0f} % ({}/{})".format(report, 100.0 * n / number_of_reports, n, number_of_reports),
                      end="\r")
                n += 1
            else:
                print("\n")

            requirements = list(filter(lambda name: name.startswith('requirement'), archive.namelist()))
            number_of_requirements = len(requirements)
            n = 1
            print("Import Requirements")
            print("===================")
            for requirement in requirements:
                with archive.open(requirement, "r") as fp:
                    file_data = json.loads(fp.read().decode())
                    json_data = {"filename": requirement.split('/')[-1],
                                 "contents": file_data["requirements"],
                                 'type': None,
                                 'date': file_data["date"]}

                    ret = self.client.post(f"/api/v1/project/{self.project}/requirements",
                                           data=json_data)
                    assert ret.status_code == 204, f"Request returned an error: {ret}"
                print("{}: {:>3.0f} % ({}/{})".format(requirement, 100.0 * n / number_of_requirements, n, number_of_requirements),
                      end="\r")
                n += 1
            else:
                print("\n")

            print("Import Tokens")
            print("=============")
            with archive.open('tokens.json', "r") as fp:
                tokens = json.loads(fp.read().decode())
                for name, token in tokens.items():
                    print("{}: {}".format(name, token), end="\r")
                    url = f"/api/v1/project/{self.project}/token"
                    json_data = {'name': name,
                                 'token': token}
                    ret = self.client.post(url, data=json_data)
                    assert ret.status_code == 201, f"Request returned an error: {ret}"
                else:
                    print("\n")

            self.rebuild_cache()

    def rebuild_cache(self):
        project_url = f"/api/v1/project/{self.project}"

        print(f"Update cache for project {self.project}")

        if self.wait:
            t0 = time.time()
            project = self.client.get(project_url)
            assert project.status_code == 200, f"Request returned an error: {project}"
            last_modified = "never"
            try:
                last_modified = project.json()['last_forced_cache_update']
            except KeyError:  # pragma: no cover
                pass  # in case no forced cache update was run before

            print(f"Wait up to {self.wait} s until the cache is updated. Last update is {last_modified}.")
        
        url = f"/api/v1/project/{self.project}/cache"
        ret = self.client.post(url, data={})
        assert ret.status_code == 204, f"Request returned an error: {ret}"

        if self.wait:
            t_max = time.time() + self.wait

            while time.time() < t_max:  # pragma: no branch
                time.sleep(1.0)
                project = self.client.get(project_url)
                assert project.status_code == 200, f"Request returned an error: {project}"
                current_last_modified = None
                try:
                    current_last_modified = project.json()['last_forced_cache_update']
                except KeyError:  # pragma: no cover
                    pass  # in case no forced cache update was run before

                if current_last_modified != last_modified:  # pragma: no branch
                    print(f"The cache was updated at {current_last_modified} within {round(time.time() - t0, 2)} s.")
                    return

            print(f"Timeout: The cache was not updated within {self.wait} s!")  # pragma: no cover
            exit(1)  # pragma: no cover
